# SPDX-License-Identifier: MIT

"""
This module focues on reading pyproject.toml fields with error collection. It is
mostly internal, except for License and Readme classes, which are re-exported in
the top-level package.
"""

from __future__ import annotations

import dataclasses
import pathlib
import re
import typing

import pdm.backend._vendor.packaging as packaging
import pdm.backend._vendor.packaging.requirements

from .errors import ErrorCollector

if typing.TYPE_CHECKING:
    from collections.abc import Generator, Iterable, Sequence

    from pdm.backend._vendor.packaging.requirements import Requirement

    from .project_table import ContactTable, Dynamic, ProjectTable


__all__ = [
    "License",
    "Readme",
]


def __dir__() -> list[str]:
    return __all__


@dataclasses.dataclass(frozen=True)
class License:
    """
    This represents a classic license, which contains text, and optionally a
    file path. Modern licenses are just SPDX identifiers, which are strings.
    """

    text: str
    file: pathlib.Path | None


@dataclasses.dataclass(frozen=True)
class Readme:
    """
    This represents a readme, which contains text and a content type, and
    optionally a file path.
    """

    text: str
    file: pathlib.Path | None
    content_type: str


T = typing.TypeVar("T")


@dataclasses.dataclass
class PyProjectReader(ErrorCollector):
    """Class for reading pyproject.toml fields with error collection.

    Unrelated errors are collected and raised at once if the `collect_errors`
    parameter is set to `True`. Some methods will return None if an error was
    raised. Most of them expect a non-None value as input to enforce the caller
    to handle missing vs. error correctly. The exact design is based on usage,
    as this is an internal class.
    """

    def ensure_str(self, value: str, key: str) -> str | None:
        """Ensure that a value is a string."""
        if isinstance(value, str):
            return value

        msg = "Field {key} has an invalid type, expecting a string"
        self.config_error(msg, key=key, got_type=type(value))
        return None

    def ensure_list(self, val: list[T], key: str) -> list[T] | None:
        """Ensure that a value is a list of strings."""
        if not isinstance(val, list):
            msg = "Field {key} has an invalid type, expecting a list of strings"
            self.config_error(msg, key=key, got_type=type(val))
            return None
        for item in val:
            if not isinstance(item, str):
                msg = "Field {key} contains item with invalid type, expecting a string"
                self.config_error(msg, key=key, got_type=type(item))
                return None

        return val

    def ensure_dict(self, val: dict[str, str], key: str) -> dict[str, str] | None:
        """Ensure that a value is a dictionary of strings."""
        if not isinstance(val, dict):
            msg = "Field {key} has an invalid type, expecting a table of strings"
            self.config_error(msg, key=key, got_type=type(val))
            return None
        for subkey, item in val.items():
            if not isinstance(item, str):
                msg = "Field {key} has an invalid type, expecting a string"
                self.config_error(msg, key=f"{key}.{subkey}", got_type=type(item))
                return None
        return val

    def ensure_people(
        self, val: Sequence[ContactTable], key: str
    ) -> list[tuple[str, str | None]]:
        """Ensure that a value is a list of tables with optional "name" and "email" keys."""
        if not isinstance(val, list):
            msg = (
                "Field {key} has an invalid type, expecting a list of "
                'tables containing the "name" and/or "email" keys'
            )
            self.config_error(msg, key=key, got_type=type(val))
            return []
        for each in val:
            if not isinstance(each, dict):
                msg = (
                    "Field {key} has an invalid type, expecting a list of "
                    'tables containing the "name" and/or "email" keys'
                    " (got list with {type_name})"
                )
                self.config_error(msg, key=key, type_name=type(each).__name__)
                return []
            for value in each.values():
                if not isinstance(value, str):
                    msg = (
                        "Field {key} has an invalid type, expecting a list of "
                        'tables containing the "name" and/or "email" keys'
                        " (got list with dict with {type_name})"
                    )
                    self.config_error(msg, key=key, type_name=type(value).__name__)
                    return []
            extra_keys = set(each) - {"name", "email"}
            if extra_keys:
                msg = (
                    "Field {key} has an invalid type, expecting a list of "
                    'tables containing the "name" and/or "email" keys'
                    " (got list with dict with extra keys {extra_keys})"
                )
                self.config_error(
                    msg,
                    key=key,
                    extra_keys=", ".join(sorted(f'"{k}"' for k in extra_keys)),
                )
                return []
        return [(entry.get("name", "Unknown"), entry.get("email")) for entry in val]

    def get_license(
        self, project: ProjectTable, project_dir: pathlib.Path
    ) -> License | str | None:
        """Get the license field from the project table. Handles PEP 639 style license too.

        None is returned if the license field is not present or if an error occurred.
        """
        val = project.get("license")
        if val is None:
            return None
        if isinstance(val, str):
            return val

        if isinstance(val, dict):
            _license = self.ensure_dict(val, "project.license")  # type: ignore[arg-type]
            if _license is None:
                return None
        else:
            msg = "Field {key} has an invalid type, expecting a string or table of strings"
            self.config_error(msg, key="project.license", got_type=type(val))
            return None

        for field in _license:
            if field not in ("file", "text"):
                msg = "Unexpected field {key}"
                self.config_error(msg, key=f"project.license.{field}")
                return None

        file: pathlib.Path | None = None
        filename = _license.get("file")
        text = _license.get("text")

        if (filename and text) or (not filename and not text):
            msg = (
                'Invalid {key} contents, expecting a string or one key "file" or "text"'
            )
            self.config_error(msg, key="project.license", got=_license)
            return None

        if filename:
            file = project_dir.joinpath(filename)
            if not file.is_file():
                msg = f"License file not found ({filename!r})"
                self.config_error(msg, key="project.license.file")
                return None
            text = file.read_text(encoding="utf-8")

        assert text is not None
        return License(text, file)

    def get_license_files(
        self, project: ProjectTable, project_dir: pathlib.Path
    ) -> list[pathlib.Path] | None:
        """Get the license-files list of files from the project table.

        Returns None if an error occurred (including invalid globs, etc) or if
        not present.
        """
        license_files = project.get("license-files")
        if license_files is None:
            return None
        if self.ensure_list(license_files, "project.license-files") is None:
            return None

        return list(self._get_files_from_globs(project_dir, license_files))

    def get_readme(  # noqa: C901
        self, project: ProjectTable, project_dir: pathlib.Path
    ) -> Readme | None:
        """Get the text of the readme from the project table.

        Returns None if an error occurred or if the readme field is not present.
        """
        if "readme" not in project:
            return None

        filename: str | None = None
        file: pathlib.Path | None = None
        text: str | None = None
        content_type: str | None = None

        readme = project["readme"]
        if isinstance(readme, str):
            # readme is a file
            text = None
            filename = readme
            if filename.endswith(".md"):
                content_type = "text/markdown"
            elif filename.endswith(".rst"):
                content_type = "text/x-rst"
            else:
                msg = "Could not infer content type for readme file {filename!r}"
                self.config_error(msg, key="project.readme", filename=filename)
                return None
        elif isinstance(readme, dict):
            # readme is a dict containing either 'file' or 'text', and content-type
            for field in readme:
                if field not in ("content-type", "file", "text"):
                    msg = "Unexpected field {key}"
                    self.config_error(msg, key=f"project.readme.{field}")
                    return None

            content_type_raw = readme.get("content-type")
            if content_type_raw is not None:
                content_type = self.ensure_str(
                    content_type_raw, "project.readme.content-type"
                )
                if content_type is None:
                    return None
            filename_raw = readme.get("file")
            if filename_raw is not None:
                filename = self.ensure_str(filename_raw, "project.readme.file")
                if filename is None:
                    return None

            text_raw = readme.get("text")
            if text_raw is not None:
                text = self.ensure_str(text_raw, "project.readme.text")
                if text is None:
                    return None

            if (filename and text) or (not filename and not text):
                msg = 'Invalid {key} contents, expecting either "file" or "text"'
                self.config_error(msg, key="project.readme", got=readme)
                return None
            if not content_type:
                msg = "Field {key} missing"
                self.config_error(msg, key="project.readme.content-type")
                return None
        else:
            msg = "Field {key} has an invalid type, expecting either a string or table of strings"
            self.config_error(msg, key="project.readme", got_type=type(readme))
            return None

        if filename:
            file = project_dir.joinpath(filename)
            if not file.is_file():
                msg = "Readme file not found ({filename!r})"
                self.config_error(msg, key="project.readme.file", filename=filename)
                return None
            text = file.read_text(encoding="utf-8")

        assert text is not None
        return Readme(text, file, content_type)

    def get_dependencies(self, project: ProjectTable) -> list[Requirement]:
        """Get the dependencies from the project table."""

        requirement_strings: list[str] | None = None
        requirement_strings_raw = project.get("dependencies")
        if requirement_strings_raw is not None:
            requirement_strings = self.ensure_list(
                requirement_strings_raw, "project.dependencies"
            )
        if requirement_strings is None:
            return []

        requirements: list[Requirement] = []
        for req in requirement_strings:
            try:
                requirements.append(packaging.requirements.Requirement(req))
            except packaging.requirements.InvalidRequirement as e:
                msg = "Field {key} contains an invalid PEP 508 requirement string {req!r} ({error!r})"
                self.config_error(msg, key="project.dependencies", req=req, error=e)
                return []
        return requirements

    def get_optional_dependencies(
        self,
        project: ProjectTable,
    ) -> dict[str, list[Requirement]]:
        """Get the optional dependencies from the project table."""

        val = project.get("optional-dependencies")
        if not val:
            return {}

        requirements_dict: dict[str, list[Requirement]] = {}
        if not isinstance(val, dict):
            msg = "Field {key} has an invalid type, expecting a table of PEP 508 requirement strings"
            self.config_error(
                msg, key="project.optional-dependencies", got_type=type(val)
            )
            return {}
        for extra, requirements in val.copy().items():
            assert isinstance(extra, str)
            if not isinstance(requirements, list):
                msg = "Field {key} has an invalid type, expecting a table of PEP 508 requirement strings"
                self.config_error(
                    msg,
                    key=f"project.optional-dependencies.{extra}",
                    got_type=type(requirements),
                )
                return {}
            requirements_dict[extra] = []
            for req in requirements:
                if not isinstance(req, str):
                    msg = "Field {key} has an invalid type, expecting a PEP 508 requirement string"
                    self.config_error(
                        msg,
                        key=f"project.optional-dependencies.{extra}",
                        got_type=type(req),
                    )
                    return {}
                try:
                    requirements_dict[extra].append(
                        packaging.requirements.Requirement(req)
                    )
                except packaging.requirements.InvalidRequirement as e:
                    msg = (
                        "Field {key} contains "
                        "an invalid PEP 508 requirement string {req!r} ({error!r})"
                    )
                    self.config_error(
                        msg,
                        key=f"project.optional-dependencies.{extra}",
                        req=req,
                        error=e,
                    )
                    return {}
        return dict(requirements_dict)

    def get_entrypoints(self, project: ProjectTable) -> dict[str, dict[str, str]]:
        """Get the entrypoints from the project table."""

        val = project.get("entry-points", None)
        if val is None:
            return {}
        if not isinstance(val, dict):
            msg = "Field {key} has an invalid type, expecting a table of entrypoint sections"
            self.config_error(msg, key="project.entry-points", got_type=type(val))
            return {}
        for section, entrypoints in val.items():
            assert isinstance(section, str)
            if not re.match(r"^\w+(\.\w+)*$", section):
                msg = (
                    "Field {key} has an invalid value, expecting a name "
                    "containing only alphanumeric, underscore, or dot characters"
                )
                self.config_error(msg, key="project.entry-points", got=section)
                return {}
            if not isinstance(entrypoints, dict):
                msg = (
                    "Field {key} has an invalid type, expecting a table of entrypoints"
                )
                self.config_error(
                    msg,
                    key=f"project.entry-points.{section}",
                    got_type=type(entrypoints),
                )
                return {}
            for name, entrypoint in entrypoints.items():
                assert isinstance(name, str)
                if not isinstance(entrypoint, str):
                    msg = "Field {key} has an invalid type, expecting a string"
                    self.config_error(
                        msg,
                        key=f"project.entry-points.{section}.{name}",
                        got_type=type(entrypoint),
                    )
                    return {}
        return val

    def get_dynamic(self, project: ProjectTable) -> list[Dynamic]:
        """Get the dynamic fields from the project table.

        Returns an empty list if the field is not present or if an error occurred.
        """
        dynamic = project.get("dynamic", [])

        self.ensure_list(dynamic, "project.dynamic")

        if "name" in dynamic:
            msg = "Unsupported field 'name' in {key}"
            self.config_error(msg, key="project.dynamic")
            return []

        return dynamic

    def _get_files_from_globs(
        self, project_dir: pathlib.Path, globs: Iterable[str]
    ) -> Generator[pathlib.Path, None, None]:
        """Given a list of globs, get files that match."""

        for glob in globs:
            if glob.startswith(("..", "/")):
                msg = "{glob!r} is an invalid {key} glob: the pattern must match files within the project directory"
                self.config_error(msg, key="project.license-files", glob=glob)
                break
            files = [f for f in project_dir.glob(glob) if f.is_file()]
            if not files:
                msg = "Every pattern in {key} must match at least one file: {glob!r} did not match any"
                self.config_error(msg, key="project.license-files", glob=glob)
                break
            for f in files:
                yield f.relative_to(project_dir)



from __future__ import annotations

import ast
import contextlib
import functools
import importlib.util
import os
import re
import sys
import types
import urllib.parse
from collections.abc import Generator, Iterable
from fnmatch import fnmatchcase
from pathlib import Path
from re import Match
from typing import Any, Callable

from pdm.backend._vendor.packaging.markers import Marker
from pdm.backend._vendor.packaging.requirements import Requirement
from pdm.backend._vendor.packaging.version import InvalidVersion, Version
from pdm.backend.exceptions import ConfigError


def safe_version(version: str) -> str:
    """
    Convert an arbitrary string to a standard version string
    """
    try:
        # normalize the version
        return str(Version(version))
    except InvalidVersion:
        version = version.replace(" ", ".")
        return re.sub("[^A-Za-z0-9.]+", "-", version)


def to_filename(name: str) -> str:
    """Convert a project or version name to its filename-escaped form

    Any '-' characters are currently replaced with '_'.
    """
    return name.replace("-", "_")


def is_python_package(fullpath: str) -> bool:
    if not os.path.isdir(fullpath):
        return False
    if os.path.basename(fullpath.rstrip("/")) in ("__pycache__", "__pypackages__"):
        return False
    if os.path.isfile(os.path.join(fullpath, "__init__.py")):
        return True
    # stubs
    return os.path.basename(fullpath).endswith("-stubs") and os.path.isfile(
        os.path.join(fullpath, "__init__.pyi")
    )


def merge_marker(requirement: Requirement, marker: str) -> None:
    """Merge the target marker str with the requirement markers"""
    if not requirement.marker:
        requirement.marker = Marker(marker)
        return
    old_marker = requirement.marker
    if "or" in old_marker._markers:
        new_marker = Marker(f"({old_marker}) and {marker}")
    else:
        new_marker = Marker(f"{old_marker} and {marker}")
    requirement.marker = new_marker


def find_packages_iter(
    where: str = ".",
    exclude: Iterable[str] = (),
    include: Iterable[str] = ("*",),
    src: str = ".",
) -> Iterable[str]:
    """
    All the packages found in 'where' that pass the 'include' filter, but
    not the 'exclude' filter.
    """

    def _build_filter(patterns: Iterable[str]) -> Callable[[str], bool]:
        """
        Given a list of patterns, return a callable that will be true only if
        the input matches at least one of the patterns.
        """
        return lambda name: any(fnmatchcase(name, pat=pat) for pat in patterns)

    fexclude, finclude = _build_filter(exclude), _build_filter(include)
    for root, dirs, _files in os.walk(where, followlinks=True):
        # Copy dirs to iterate over it, then empty dirs.
        all_dirs = dirs[:]
        dirs[:] = []

        for dir in all_dirs:
            full_path = os.path.join(root, dir)
            rel_path = os.path.relpath(full_path, src)
            package = rel_path.replace(os.path.sep, ".")
            # Skip directory trees that are not valid packages
            if "." in dir:
                continue

            # Should this package be included?
            if (
                os.path.isfile(os.path.join(full_path, "__init__.py"))
                and finclude(package)
                and not fexclude(package)
            ):
                yield package

            # Keep searching subdirectories, as there may be more packages
            # down there, even if the parent was excluded.
            dirs.append(dir)


def normalize_path(filename: str | Path) -> str:
    """Normalize a file/dir name for comparison purposes"""
    filename = os.path.abspath(filename) if sys.platform == "cygwin" else filename
    return os.path.normcase(os.path.realpath(os.path.normpath(filename)))


def is_relative_path(target: Path, other: Path) -> bool:
    try:
        target.relative_to(other)
    except ValueError:
        return False
    else:
        return True


def expand_vars(line: str, root: str) -> str:
    """Expand environment variables in a string."""
    if "$" not in line:
        return line

    if "://" in line:
        quote: Callable[[str], str] = urllib.parse.quote
    else:
        quote = str
    line = line.replace("file:///${PROJECT_ROOT}", Path(root).as_uri())

    def replace_func(match: Match[str]) -> str:
        rv = os.getenv(match.group(1))
        if rv is None:
            return match.group(0)
        return quote(rv)

    return re.sub(r"\$\{(.+?)\}", replace_func, line)


def import_module_at_path(
    src_path: str | Path, module_name: str = "_local", context: Path | None = None
) -> types.ModuleType:
    """Import a module from a given path."""
    spec = importlib.util.spec_from_file_location(module_name, src_path)
    if spec is None:
        raise ValueError(f"Could not import module {module_name} from {src_path}")
    if context is not None:
        sys.path.insert(0, str(context.absolute()))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)  # type: ignore
    if context is not None:
        sys.path.pop(0)
    return module


def normalize_file_permissions(st_mode: int) -> int:
    """
    Normalizes the permission bits in the st_mode field from stat to 644/755
    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalising
    to 644 (non executable) or 755 (executable) makes builds more reproducible.
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:
        new_mode |= 0o111  # Executable: 644 -> 755

    return new_mode


@contextlib.contextmanager
def patch_sys_path(path: str | Path) -> Generator[None]:
    old_path = sys.path[:]
    sys.path.insert(0, str(path))
    try:
        yield
    finally:
        sys.path[:] = old_path


_attr_regex = re.compile(r"([\w.]+):([\w.]+)\s*(\([^)]+\))?")


def evaluate_module_attribute(
    expression: str, context: Path | None = None
) -> tuple[Any, tuple[Any, ...]]:
    """Evaluate the value of an expression like '<module>:<attribute>'

    Returns:
        the object and the calling arguments if any
    """
    if context is None:
        cm = contextlib.nullcontext()
    else:
        cm = patch_sys_path(context)  # type: ignore[assignment]

    matched = _attr_regex.match(expression)
    if matched is None:
        raise ConfigError(
            "Invalid expression, must be in the format of `module:attribute`."
        )
    with cm:
        module = importlib.import_module(matched.group(1))
        attrs = matched.group(2).split(".")
        obj: Any = functools.reduce(getattr, attrs, module)
        args_group = matched.group(3)
        if args_group:
            # make tuple
            args_group = args_group.strip()[:-1] + ",)"
            args = ast.literal_eval(args_group)

        else:
            args = ()
        return obj, args



from __future__ import annotations


class BuildError(RuntimeError):
    pass


class ConfigError(ValueError):
    pass


class PDMWarning(UserWarning):
    pass


class ValidationError(ConfigError):
    def __init__(self, summary: str, key: str | None = None) -> None:
        super().__init__(summary)
        self.key = key

    def __str__(self) -> str:
        prefix = f"{self.key}: " if self.key else ""
        return f"{prefix}{self.args[0]}"



# SPDX-License-Identifier: MIT

"""
This module contains type definitions for the tables used in the
``pyproject.toml``.  You should either import this at type-check time only, or
make sure ``typing_extensions`` is available for Python 3.10 and below.

Documentation notice: the fields with hyphens are not shown due to a sphinx-autodoc bug.
"""

from __future__ import annotations

import sys
import typing
from typing import Any, Dict, List, Union

if sys.version_info < (3, 11):
    from typing_extensions import Required
else:
    from typing import Required

if sys.version_info < (3, 8):
    from typing_extensions import Literal, TypedDict
else:
    from typing import Literal, TypedDict


__all__ = [
    "BuildSystemTable",
    "ContactTable",
    "Dynamic",
    "IncludeGroupTable",
    "LicenseTable",
    "ProjectTable",
    "PyProjectTable",
    "ReadmeTable",
]


def __dir__() -> list[str]:
    return __all__


class ContactTable(TypedDict, total=False):
    name: str
    email: str


class LicenseTable(TypedDict, total=False):
    text: str
    file: str


ReadmeTable = TypedDict(
    "ReadmeTable", {"file": str, "text": str, "content-type": str}, total=False
)

Dynamic = Literal[
    "authors",
    "classifiers",
    "dependencies",
    "description",
    "dynamic",
    "entry-points",
    "gui-scripts",
    "keywords",
    "license",
    "maintainers",
    "optional-dependencies",
    "readme",
    "requires-python",
    "scripts",
    "urls",
    "version",
]

ProjectTable = TypedDict(
    "ProjectTable",
    {
        "name": Required[str],
        "version": str,
        "description": str,
        "license": Union[LicenseTable, str],
        "license-files": List[str],
        "readme": Union[str, ReadmeTable],
        "requires-python": str,
        "dependencies": List[str],
        "optional-dependencies": Dict[str, List[str]],
        "entry-points": Dict[str, Dict[str, str]],
        "authors": List[ContactTable],
        "maintainers": List[ContactTable],
        "urls": Dict[str, str],
        "classifiers": List[str],
        "keywords": List[str],
        "scripts": Dict[str, str],
        "gui-scripts": Dict[str, str],
        "dynamic": List[Dynamic],
    },
    total=False,
)

BuildSystemTable = TypedDict(
    "BuildSystemTable",
    {
        "build-backend": str,
        "requires": List[str],
        "backend-path": List[str],
    },
    total=False,
)

# total=False here because this could be
# extended in the future
IncludeGroupTable = TypedDict(
    "IncludeGroupTable",
    {"include-group": str},
    total=False,
)

PyProjectTable = TypedDict(
    "PyProjectTable",
    {
        "build-system": BuildSystemTable,
        "project": ProjectTable,
        "tool": Dict[str, Any],
        "dependency-groups": Dict[str, List[Union[str, IncludeGroupTable]]],
    },
    total=False,
)

# Tests for type checking
if typing.TYPE_CHECKING:
    PyProjectTable(
        {
            "build-system": BuildSystemTable(
                {"build-backend": "one", "requires": ["two"]}
            ),
            "project": ProjectTable(
                {
                    "name": "one",
                    "version": "0.1.0",
                }
            ),
            "tool": {"thing": object()},
            "dependency-groups": {
                "one": [
                    "one",
                    IncludeGroupTable({"include-group": "two"}),
                ]
            },
        }
    )



from __future__ import annotations

import glob
import os
import shutil
import sys
from collections.abc import Iterable, Mapping
from fnmatch import fnmatch
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, TypeVar, cast

from pdm.backend._vendor.pyproject_metadata import StandardMetadata
from pdm.backend.config import Config
from pdm.backend.hooks import BuildHookInterface, Context
from pdm.backend.hooks.version import DynamicVersionBuildHook
from pdm.backend.structures import FileMap
from pdm.backend.utils import expand_vars, import_module_at_path, is_python_package

if TYPE_CHECKING:
    from typing import SupportsIndex

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points

T = TypeVar("T", bound="Builder")


def is_same_or_descendant_path(target: str, path: str) -> bool:
    """Check target is same or descendant with path"""
    try:
        Path(target).relative_to(path)
        return True
    except ValueError:
        return False


def _merge_globs(
    include_globs: dict[str, str], excludes_globs: dict[str, str]
) -> tuple[list[str], list[str]]:
    """Correctly merge includes and excludes.
    When a pattern exists in both includes and excludes,
    determine the priority in the following ways:

    1. The one with more parts in the path wins
    2. If part numbers are equal, the one that is more concrete wins
    3. If both have the same part number and concrete level, *excludes* wins
    """

    def path_weight(pathname: str) -> tuple[int, int]:
        """Return a two-element tuple [part_num, concrete_level]"""
        pathname_parts = Path(pathname).parts
        wildcard_count = 0
        if glob.has_magic(pathname):
            for part in pathname_parts:
                if part == "**":
                    wildcard_count += 2
                elif glob.has_magic(part):
                    wildcard_count += 1
        # The concrete level is the opposite of wildcard_count
        return len(pathname_parts), -wildcard_count

    includes = []
    for path, key in include_globs.items():
        if path in excludes_globs:
            if path_weight(key) <= path_weight(excludes_globs[path]):
                continue  # Exclude wins
            else:
                del excludes_globs[path]  # Include wins
        includes.append(path)
    return includes, list(excludes_globs)


def _find_top_packages(root: str) -> list[str]:
    result = []
    for path in os.listdir(root):
        path = os.path.join(root, path)
        if is_python_package(path):
            result.append(path)
    return result


Target = Literal["sdist", "wheel", "editable"]


class Builder:
    """Base class for building and distributing a package from given path."""

    DEFAULT_EXCLUDES = [".pdm-build"]

    target: Target
    hooks: list[BuildHookInterface] = [DynamicVersionBuildHook()]

    def __init__(
        self,
        location: str | Path,
        config_settings: Mapping[str, Any] | None = None,
    ) -> None:
        self._old_cwd: str | None = None
        self.location = Path(location)
        self.config = Config.from_pyproject(self.location)
        self.config_settings = dict(config_settings or {})
        self._hooks = list(self.get_hooks())

    def __reduce_ex__(self, __protocol: SupportsIndex = 3) -> str | tuple[Any, ...]:
        return (
            self.__class__,
            (self.location, self.config_settings),
            {"config": self.config},
        )

    def get_hooks(self) -> Iterable[BuildHookInterface]:
        """Load hooks in the following order:

        1. plugins installed in 'pdm.build.hook' entry point group.
        2. local hook defined in the `pdm_build.py` file.
        """
        yield from self.hooks
        for ep in entry_points(group="pdm.build.hook"):
            hook = ep.load()
            try:
                yield cast(BuildHookInterface, hook())  # for a hook class
            except TypeError:
                yield cast(BuildHookInterface, hook)  # for a hook module
        local_hook = self.config.build_config.custom_hook
        if local_hook is not None:
            yield cast(
                BuildHookInterface,
                import_module_at_path(
                    self.location / local_hook, context=self.location
                ),
            )

    def call_hook(
        self, hook_name: str, context: Context, *args: Any, **kwargs: Any
    ) -> None:
        """Call the hook on all registered hooks and skip if not implemented."""
        for hook in self._hooks:
            if hasattr(hook, "pdm_build_hook_enabled"):
                if not hook.pdm_build_hook_enabled(context):
                    continue
            if hasattr(hook, hook_name):
                getattr(hook, hook_name)(context, *args, **kwargs)

    def build_context(self, destination: Path, **kwargs: Any) -> Context:
        build_dir = self.location / ".pdm-build"
        if not destination.exists():
            destination.mkdir(0o700, parents=True)
        return Context(
            build_dir=build_dir, dist_dir=destination, kwargs=kwargs, builder=self
        )

    def __enter__(self: T) -> T:
        self._old_cwd = os.getcwd()
        os.chdir(self.location)
        return self

    def __exit__(self, *args: Any) -> None:
        assert self._old_cwd
        os.chdir(self._old_cwd)

    def clean(self, context: Context) -> None:
        """Clean up the build directory."""
        self.call_hook("pdm_build_clean", context)
        if context.build_dir.exists():
            shutil.rmtree(context.build_dir)

    def _fix_dependencies(self) -> None:
        """Fix the dependencies and remove dynamic variables from the metadata"""
        metadata = self.config.metadata
        root = self.location.as_posix()
        if metadata.get("dependencies"):
            metadata["dependencies"] = [
                expand_vars(dep, root) for dep in metadata["dependencies"]
            ]
        if metadata.get("optional-dependencies"):
            for name, deps in metadata["optional-dependencies"].items():
                metadata["optional-dependencies"][name] = [
                    expand_vars(dep, root) for dep in deps
                ]

    def initialize(self, context: Context) -> None:
        """Initialize the build context."""
        self._fix_dependencies()
        self.call_hook("pdm_build_initialize", context)

    def get_files(self, context: Context) -> Iterable[tuple[str, Path]]:
        """Get the files to add to the package, return a iterable of
        (relpath, path).
        """
        files = self._collect_files(context)
        self.call_hook("pdm_build_update_files", context, files)
        # At this point, all files must be ready under the build_dir,
        # collect them now.
        files.update(self._collect_build_files(context))
        return files.items()

    def finalize(self, context: Context, artifact: Path) -> None:
        self.call_hook("pdm_build_finalize", context, artifact)

    def build(self, build_dir: str, **kwargs: Any) -> Path:
        """Build the package and return the path to the artifact."""
        context = self.build_context(Path(build_dir), **kwargs)
        should_clean = True

        if "no-clean-build" in self.config_settings:
            should_clean = False
        elif "PDM_BUILD_NO_CLEAN" in os.environ:
            should_clean = os.getenv("PDM_BUILD_NO_CLEAN", "0").lower() in (
                "0",
                "false",
                "no",
            )

        if should_clean:
            self.clean(context)

        self.initialize(context)
        files = sorted(self.get_files(context))
        artifact = self.build_artifact(context, files)
        self.finalize(context, artifact)
        return artifact

    def build_artifact(
        self, context: Context, files: Iterable[tuple[str, Path]]
    ) -> Path:
        """Build the artifact from an iterable of (relpath, path) pairs
        and return the path to it.
        """
        raise NotImplementedError()

    def _collect_files(self, context: Context) -> FileMap:
        """Collect files to add to the artifact under the given root."""
        root = self.location
        includes, excludes = self._get_include_and_exclude_paths()
        files = FileMap()
        for include_path in includes:
            path = root / include_path
            if path.is_file():
                files[include_path] = path
                continue
            # The path is a directory name
            for p in path.glob("**/*"):
                if not p.is_file():
                    continue

                rel_path = p.absolute().relative_to(root).as_posix()
                if p.name.endswith(".pyc") or self._is_excluded(rel_path, excludes):
                    continue

                files[rel_path] = p

        return files

    def _collect_build_files(self, context: Context) -> FileMap:
        """Collect files from the build directory."""
        files = FileMap()
        if not context.build_dir.exists():
            return files
        _, excludes = self._get_include_and_exclude()
        for p in context.build_dir.glob("**/*"):
            if not p.is_file():
                continue
            rel_path = p.absolute().relative_to(context.build_dir).as_posix()
            if not self._is_excluded(rel_path, excludes):
                files[rel_path] = p
        return files

    def find_license_files(self, metadata: StandardMetadata) -> list[str]:
        result: list[str] = []
        if file := getattr(metadata.license, "file", None):
            result.append(file.relative_to(self.location).as_posix())
        if metadata.license_files:
            for file in metadata.license_files:
                result.append(file.as_posix())
        if (
            not result and metadata.license_files is None
        ):  # no license files specified, find from default patterns for backward compatibility
            for pattern in ["LICEN[CS]E*", "COPYING*", "NOTICE*"]:
                for path in self.location.glob(pattern):
                    if path.is_file():
                        result.append(path.relative_to(self.location).as_posix())
        return result

    def _get_include_and_exclude(self) -> tuple[set[str], set[str]]:
        includes = set()
        excludes = set(self.DEFAULT_EXCLUDES)
        build_config = self.config.build_config
        meta_excludes = list(build_config.excludes)
        source_includes = build_config.source_includes or ["tests"]
        if self.target != "sdist":
            # exclude source-includes for non-sdist builds
            meta_excludes.extend(source_includes)

        if not build_config.includes:
            top_packages = _find_top_packages(build_config.package_dir or ".")
            if top_packages:
                includes.update(top_packages)
            else:
                # Include all top-level .py modules under the package-dir
                includes.add(f"{build_config.package_dir or '.'}/*.py")
        else:
            # use what user specifies
            includes.update(build_config.includes)

        includes.update(source_includes)
        excludes.update(meta_excludes)
        return includes, excludes

    def _get_include_and_exclude_paths(self) -> tuple[list[str], list[str]]:
        includes, excludes = self._get_include_and_exclude()
        include_globs = {
            os.path.normpath(path): key
            for key in includes
            for path in glob.iglob(key, recursive=True)
        }

        excludes_globs = {
            os.path.normpath(path): key
            for key in excludes
            for path in glob.iglob(key, recursive=True)
        }

        include_paths, exclude_paths = _merge_globs(include_globs, excludes_globs)
        return sorted(include_paths), sorted(exclude_paths)

    def _is_excluded(self, path: str, exclude_paths: Iterable[str]) -> bool:
        return any(
            is_same_or_descendant_path(path, exclude_path)
            or fnmatch(path, exclude_path)
            for exclude_path in exclude_paths
        )

    def _show_add_file(self, rel_path: str, full_path: Path) -> None:
        try:
            show_path = full_path.relative_to(self.location)
        except ValueError:
            show_path = full_path
        print(f" - Adding {show_path} -> {rel_path}")



"""
module to get version from tag of SCM repository.
Adapted from setuptools-scm. Currently only support git and hg.
"""

from __future__ import annotations

import os
import re
import shlex
import shutil
import subprocess
import warnings
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

from pdm.backend._vendor.packaging.version import Version

if TYPE_CHECKING:
    from _typeshed import StrPath

DEFAULT_TAG_REGEX = re.compile(
    r"^(?:[\w-]+-)?(?P<version>[vV]?\d+(?:\.\d+){0,2}[^\+]*)(?:\+.*)?$"
)


@dataclass(frozen=True)
class Config:
    tag_regex: re.Pattern
    tag_filter: str | None


def _subprocess_call(
    cmd: str | list[str],
    cwd: StrPath | None = None,
    extra_env: dict[str, str] | None = None,
) -> tuple[int, str, str]:
    # adapted from pre-commit
    # Too many bugs dealing with environment variables and GIT:
    # https://github.com/pre-commit/pre-commit/issues/300
    env = {
        k: v
        for k, v in os.environ.items()
        if not k.startswith("GIT_")
        or k in ("GIT_EXEC_PATH", "GIT_SSH", "GIT_SSH_COMMAND")
    }
    # adapted from setuptools-scm
    # PYTHONPATH can break Python applications installed with Pipx, UV, etc.
    if "PYTHONPATH" in env:
        env["PYTHONPATH"] = os.pathsep.join(
            path
            for path in env["PYTHONPATH"].split(os.pathsep)
            if "-build-env-" not in path
        )
    env.update({"LC_ALL": "C", "LANG": "", "HGPLAIN": "1"})
    if extra_env:
        env.update(extra_env)
    if isinstance(cmd, str):
        cmd = shlex.split(cmd)
    try:
        proc = subprocess.Popen(
            cmd, cwd=cwd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = proc.communicate()
    except NotADirectoryError as e:
        if hasattr(e, "winerror") and e.winerror == 267:
            # https://github.com/pdm-project/pdm-backend/issues/197
            # More helpful diagnostic.
            import textwrap

            err_msg = f"""\
                       The above error is most likely caused by an unsupported
                       MSYS2 git binary at {cmd[0]}. Please point your PATH to
                       a different git binary such as Git For Windows.
                       """
            raise subprocess.SubprocessError(textwrap.dedent(err_msg)) from e
        else:
            raise e
    return (
        proc.returncode,
        out.decode("utf-8", "surrogateescape").strip(),
        err.decode("utf-8", "surrogateescape").strip(),
    )


class SCMVersion(NamedTuple):
    version: Version
    distance: int | None
    dirty: bool
    node: str | None
    branch: str | None


def meta(
    config: Config,
    tag: str | Version,
    distance: int | None = None,
    dirty: bool = False,
    node: str | None = None,
    branch: str | None = None,
) -> SCMVersion:
    if isinstance(tag, str):
        tag = tag_to_version(config, tag)
    return SCMVersion(tag, distance, dirty, node, branch)


def _git_get_branch(root: StrPath) -> str | None:
    ret, out, _ = _subprocess_call("git rev-parse --abbrev-ref HEAD", root)
    if not ret:
        return out
    return None


def _git_is_dirty(root: StrPath) -> bool:
    _, out, _ = _subprocess_call("git status --porcelain --untracked-files=no", root)
    return bool(out)


def _git_get_node(root: StrPath) -> str | None:
    ret, out, _ = _subprocess_call("git rev-parse --verify --quiet HEAD", root)
    if not ret:
        return out[:7]
    return None


def _git_count_all_nodes(root: StrPath) -> int:
    _, out, _ = _subprocess_call("git rev-list HEAD", root)
    return out.count("\n") + 1


def _git_parse_describe(describe_output: str) -> tuple[str, int, str, bool]:
    # 'describe_output' looks e.g. like 'v1.5.0-0-g4060507' or
    # 'v1.15.1rc1-37-g9bd1298-dirty'.

    if describe_output.endswith("-dirty"):
        dirty = True
        describe_output = describe_output[:-6]
    else:
        dirty = False

    tag, number, node = describe_output.rsplit("-", 2)
    return tag, int(number), node, dirty


class _ParseResult(NamedTuple):
    version: str
    prefix: str
    suffix: str


def _parse_version_tag(config: Config, tag: str) -> _ParseResult | None:
    tagstring = tag if not isinstance(tag, str) else str(tag)
    match = config.tag_regex.match(tagstring)

    result = None
    if match:
        if len(match.groups()) == 1:
            key: int | str = 1
        else:
            key = "version"

        result = _ParseResult(
            match.group(key),
            match.group(0)[: match.start(key)],
            match.group(0)[match.end(key) :],
        )

    return result


def tag_to_version(config: Config, tag: str) -> Version:
    """
    take a tag that might be prefixed with a keyword and return only the version part
    :param config: optional configuration object
    """
    tagdict = _parse_version_tag(config, tag)
    if not tagdict or not tagdict.version:
        warnings.warn(f"tag {tag!r} no version found")
        return Version("0.0.0")

    version = tagdict.version

    if tagdict.suffix:
        warnings.warn(f"tag {tag!r} will be stripped of its suffix {tagdict.suffix!r}")

    return Version(version)


def git_parse_version(root: StrPath, config: Config) -> SCMVersion | None:
    git = shutil.which("git")
    if not git:
        return None

    ret, repo, _ = _subprocess_call([git, "rev-parse", "--show-toplevel"], root)
    if ret or not repo:
        return None

    if os.path.isfile(os.path.join(repo, ".git/shallow")):
        warnings.warn(f"{repo!r} is shallow and may cause errors")
    describe_cmd = [
        git,
        "-c",
        "core.abbrev=auto",
        "describe",
        "--dirty",
        "--tags",
        "--long",
        "--match",
        config.tag_filter or "*.*",
    ]
    ret, output, _ = _subprocess_call(describe_cmd, repo)
    branch = _git_get_branch(repo)

    if ret:
        rev_node = _git_get_node(repo)
        dirty = _git_is_dirty(repo)
        if rev_node is None:
            return meta(config, "0.0", 0, dirty)
        return meta(
            config, "0.0", _git_count_all_nodes(repo), dirty, f"g{rev_node}", branch
        )
    else:
        tag, number, node, dirty = _git_parse_describe(output)
        return meta(config, tag, number or None, dirty, node, branch)


def get_distance_revset(tag: str | None) -> str:
    return (
        "(branch(.)"  # look for revisions in this branch only
        " and {rev}::."  # after the last tag
        # ignore commits that only modify .hgtags and nothing else:
        " and (merge() or file('re:^(?!\\.hgtags).*$'))"
        " and not {rev})"  # ignore the tagged commit itself
    ).format(rev=f"tag({tag!r})" if tag is not None else "null")


def hg_get_graph_distance(root: StrPath, tag: str | None) -> int:
    cmd = ["hg", "log", "-q", "-r", get_distance_revset(tag)]
    _, out, _ = _subprocess_call(cmd, root)
    return len(out.strip().splitlines())


def _hg_tagdist_normalize_tagcommit(
    config: Config,
    tag: str,
    dist: int,
    node: str,
    branch: str,
    dirty: bool,
) -> SCMVersion:
    return meta(
        config, tag, distance=dist or None, node=node, dirty=dirty, branch=branch
    )


def guess_next_version(tag_version: Version) -> str:
    version = _strip_local(str(tag_version))
    return _bump_dev(version) or _bump_regex(version)


def _strip_local(version_string: str) -> str:
    public, _, _ = version_string.partition("+")
    return public


def _bump_dev(version: str) -> str:
    if ".dev" not in version:
        return ""

    prefix, tail = version.rsplit(".dev", 1)
    assert tail == "0", "own dev numbers are unsupported"
    return prefix


def _bump_regex(version: str) -> str:
    match = re.match(r"(.*?)(\d+)$", version)
    assert match is not None
    prefix, tail = match.groups()
    return f"{prefix}{int(tail) + 1}"


def hg_parse_version(root: StrPath, config: Config) -> SCMVersion | None:
    hg = shutil.which("hg")
    if not hg:
        return None

    tag_filter = config.tag_filter or "\\."
    _, output, _ = _subprocess_call(
        [
            hg,
            "log",
            "-r",
            ".",
            "--template",
            f"{{latesttag(r're:{tag_filter}')}}-{{node|short}}-{{branch}}",
        ],
        root,
    )
    tag: str | None
    try:
        tag, node, branch = output.rsplit("-", 2)
    except ValueError:
        return None  # unpacking failed, unexpected hg repo
    # If no tag exists passes the tag filter.
    if tag == "null":
        tag = None

    _, id_output, _ = _subprocess_call(
        [hg, "id", "-i"],
        root,
    )
    dirty = id_output.endswith("+")
    try:
        dist = hg_get_graph_distance(root, tag)
        if tag is None:
            tag = "0.0"
        return _hg_tagdist_normalize_tagcommit(
            config, tag, dist, node, branch, dirty=dirty
        )
    except ValueError:
        return None  # unpacking failed, old hg


def default_version_formatter(version: SCMVersion) -> str:
    if version.distance is None:
        main_version = str(version.version)
    else:
        guessed = guess_next_version(version.version)
        main_version = f"{guessed}.dev{version.distance}"

    if version.distance is None or version.node is None:
        clean_format = ""
        dirty_format = "+d{time:%Y%m%d}"
    else:
        clean_format = "+{node}"
        dirty_format = "+{node}.d{time:%Y%m%d}"
    fmt = dirty_format if version.dirty else clean_format
    local_version = fmt.format(node=version.node, time=datetime.now(tz=timezone.utc))
    return main_version + local_version


def get_version_from_scm(
    root: str | Path, *, tag_regex: str | None = None, tag_filter: str | None = None
) -> SCMVersion | None:
    config = Config(
        tag_regex=re.compile(tag_regex) if tag_regex else DEFAULT_TAG_REGEX,
        tag_filter=tag_filter,
    )
    for func in (git_parse_version, hg_parse_version):
        version = func(root, config)
        if version:
            return version
    return None
